@@ -23,16 +23,40 @@
 #ifndef MODUTILS_UTIL_H
 #define MODUTILS_UTIL_H 1
 
-#ident "$Id: util.h,v 1.1 1997/09/10 22:13:04 rth Exp $"
+#ident "$Id: util.h 1.1 Tue, 24 Aug 1999 23:26:49 -0700 keith $"
 
-void error(const char *, ...)
-#ifdef __GNUC__
-  __attribute__((format(printf, 1, 2)))
-#endif
-  ;
+#define SHELL_META "&();|<>$`\"'\\!{}[]~=+:?*" /* Sum of bj0rn and Debian */
 
 void *xmalloc(size_t);
 void *xrealloc(void *, size_t);
 char *xstrdup(const char *);
 
+/* Error logging */
+extern int log;
+extern int errors;
+extern const char *error_file;
+
+void error(const char *fmt, ...)
+#ifdef __GNUC__
+  __attribute__((format(printf, 1, 2)))
+#endif
+  ;
+
+void lprintf(const char *fmt, ...)
+#ifdef __GNUC__
+  __attribute__((format(printf, 1, 2)))
+#endif
+  ;
+
+void setsyslog(const char *program);
+
+/*
+ * Generic globlist <bj0rn@blox.se>
+ */
+typedef struct {
+	int pathc;       /* Count of paths matched so far  */
+	char **pathv;    /* List of matched pathnames.  */
+} GLOB_LIST;
+int meta_expand(char *pt, GLOB_LIST *g, char *base_dir, char *version);
+
 #endif /* util.h */
