@@ -19,6 +19,7 @@
    along with this program; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
 
+#ident "$Id: obj_reloc.c 1.1 Tue, 24 Aug 1999 23:26:49 -0700 keith $"
 
 #include <string.h>
 #include <assert.h>
@@ -78,23 +79,17 @@ obj_symbol_patch(struct obj_file *f, int secidx, ElfW(Addr) offset,
 }
 
 int
-obj_allocate_commons_and_check_undefineds(struct obj_file *f)
+obj_check_undefineds(struct obj_file *f, int quiet)
 {
-  struct common_entry
-  {
-    struct common_entry *next;
-    struct obj_symbol *sym;
-  } *common_head = NULL;
-
-  int i, ret = 1;
+  unsigned long i;
+  int ret = 1;
 
   for (i = 0; i < HASH_BUCKETS; ++i)
     {
       struct obj_symbol *sym;
       for (sym = f->symtab[i]; sym ; sym = sym->next)
-	switch (sym->secidx)
+	if (sym->secidx == SHN_UNDEF)
 	  {
-	  case SHN_UNDEF:
 	    if (ELFW(ST_BIND)(sym->info) == STB_WEAK)
 	      {
 		sym->secidx = SHN_ABS;
@@ -102,12 +97,33 @@ obj_allocate_commons_and_check_undefineds(struct obj_file *f)
 	      }
 	    else
 	      {
-		error("unresolved symbol %s", sym->name);
+		if (!quiet)
+			error("unresolved symbol %s", sym->name);
 		ret = 0;
 	      }
-	    break;
+	  }
+    }
 
-	  case SHN_COMMON:
+  return ret;
+}
+
+void
+obj_allocate_commons(struct obj_file *f)
+{
+  struct common_entry
+  {
+    struct common_entry *next;
+    struct obj_symbol *sym;
+  } *common_head = NULL;
+
+  unsigned long i;
+
+  for (i = 0; i < HASH_BUCKETS; ++i)
+    {
+      struct obj_symbol *sym;
+      for (sym = f->symtab[i]; sym ; sym = sym->next)
+	if (sym->secidx == SHN_COMMON)
+	  {
 	    /* Collect all COMMON symbols and sort them by size so as to
 	       minimize space wasted by alignment requirements.  */
 	    {
@@ -121,12 +137,28 @@ obj_allocate_commons_and_check_undefineds(struct obj_file *f)
 	      n->sym = sym;
 	      *p = n;
 	    }
-	  break;
 	  }
     }
 
-  if (!ret)
-    return ret;
+  for (i = 1; i < f->local_symtab_size; ++i)
+    {
+      struct obj_symbol *sym = f->local_symtab[i];
+      if (sym && sym->secidx == SHN_COMMON)
+	{
+	  struct common_entry **p, *n;
+	  for (p = &common_head; *p ; p = &(*p)->next)
+	    if (sym == (*p)->sym)
+	      break;
+	    else if (sym->size < (*p)->sym->size)
+	      {
+		n = alloca(sizeof(*n));
+		n->next = *p;
+		n->sym = sym;
+		*p = n;
+		break;
+	      }
+	}
+    }
 
   if (common_head)
     {
@@ -145,7 +177,7 @@ obj_allocate_commons_and_check_undefineds(struct obj_file *f)
 	  f->header.e_shnum = i+1;
 
 	  memset(sec, 0, sizeof(*sec));
-	  sec->header.sh_type = SHT_NOBITS;
+	  sec->header.sh_type = SHT_PROGBITS;
 	  sec->header.sh_flags = SHF_WRITE|SHF_ALLOC;
 	  sec->name = ".bss";
 	  sec->idx = i;
@@ -187,10 +219,9 @@ obj_allocate_commons_and_check_undefineds(struct obj_file *f)
 	{
 	  s->contents = memset(xmalloc(s->header.sh_size),
 			       0, s->header.sh_size);
+	  s->header.sh_type = SHT_PROGBITS;
 	}
     }
-
-  return ret;
 }
 
 unsigned long
@@ -264,19 +295,35 @@ obj_relocate (struct obj_file *f, ElfW(Addr) base)
 	  if (symndx)
 	    {
 	      /* Note we've already checked for undefined symbols.  */
-	      const char *name;
 
 	      extsym = &symtab[symndx];
-	      if (extsym->st_name)
-		name = strtab + extsym->st_name;
+	      if (ELFW(ST_BIND)(extsym->st_info) == STB_LOCAL)
+		{
+		  /* Local symbols we look up in the local table to be sure
+		     we get the one that is really intended.  */
+		  intsym = f->local_symtab[symndx];
+		}
 	      else
-		name = f->sections[extsym->st_shndx]->name;
+		{
+		  /* Others we look up in the hash table.  */
+	          const char *name;
+		  if (extsym->st_name)
+		    name = strtab + extsym->st_name;
+	          else
+		    name = f->sections[extsym->st_shndx]->name;
+	          intsym = obj_find_symbol(f, name);
+		}
 
-	      intsym = obj_find_symbol(f, name);
 	      value = obj_symbol_final_value(f, intsym);
 	      intsym->referenced = 1;
 	    }
+
 #if SHT_RELM == SHT_RELA
+#if defined(__alpha__) && defined(AXP_BROKEN_GAS)
+          /* Work around a nasty GAS bug, that is fixed as of 2.7.0.9.  */
+          if (!extsym || !extsym->st_name ||
+              ELFW(ST_BIND)(extsym->st_info) != STB_LOCAL)
+#endif
 	  value += rel->r_addend;
 #endif
 
@@ -325,13 +372,6 @@ obj_relocate (struct obj_file *f, ElfW(Addr) base)
       for (p = f->string_patches; p ; p = p->next)
 	{
 	  struct obj_section *targsec = f->sections[p->reloc_secidx];
-	  if (targsec->header.sh_type == SHT_NOBITS)
-	    {
-	      /* Convert the section to PROGBITS so we can write to it.  */
-	      targsec->contents = memset(xmalloc(targsec->header.sh_size),
-					 0, targsec->header.sh_size);
-	      targsec->header.sh_type = SHT_PROGBITS;
-	    }
 	  *(ElfW(Addr) *)(targsec->contents + p->reloc_offset)
 	    = strsec_base + p->string_offset;
 	}
@@ -344,13 +384,6 @@ obj_relocate (struct obj_file *f, ElfW(Addr) base)
       for (p = f->symbol_patches; p; p = p->next)
 	{
 	  struct obj_section *targsec = f->sections[p->reloc_secidx];
-	  if (targsec->header.sh_type == SHT_NOBITS)
-	    {
-	      /* Convert the section to PROGBITS so we can write to it.  */
-	      targsec->contents = memset(xmalloc(targsec->header.sh_size),
-					 0, targsec->header.sh_size);
-	      targsec->header.sh_type = SHT_PROGBITS;
-	    }
 	  *(ElfW(Addr) *)(targsec->contents + p->reloc_offset)
 	    = obj_symbol_final_value(f, p->sym);
 	}
@@ -369,7 +402,7 @@ obj_create_image (struct obj_file *f, char *image)
     {
       char *secimg;
 
-      if (sec->header.sh_size == 0)
+      if (sec->contents == 0)
 	continue;
 
       secimg = image + (sec->header.sh_addr - base);
